import numpy as np
from .boson_waveform import BosonCloudWaveform
from .cloud_model import CloudModel 
from .harmonics import sYlm 
from .units import set_units
import warnings

class MatchedWaveform(BosonCloudWaveform):
    def __init__(self, mu, Mbh0, abh0, cloud_model, units="natural"):
        """
        Calculate some derived quantites using the specified CloudModel
        for the specified parameters:
        mu : ultralight boson mass
        Mbh0 : initial black hole mass (before cloud growth)
        abh0 : initial black hole dimensionless (before cloud growth)
        
        Internally G=c=1 and mu has dimensions of 1/mass
        
        See units.py for input/output units.

        If "+alpha" is appended to either "physical" or "natural," then units
        are the same as above, except the input mu is taken to be in units of 
        (hbar c^3)/(G*Mbh0), i.e.  mu is set to the dimensionless 
        "fine structure constant" alpha. 
        """
        if not isinstance(cloud_model, CloudModel):
            raise TypeError
        if (mu<0 or Mbh0<0 or abh0<=0 or abh0>cloud_model.max_spin()):
            raise ValueError("Invalid boson cloud waveform parameters: mu,Mbh0,abh0<=0, or abh0>max. spin the cloud_model is valid for.")

        #Set units
        (self._tunit, self._Punit, self._dunit, self._hbar, mu_fac) = set_units(units, Mbh0)
        mu = mu_fac*mu

        #Check if superradiant condition is met for m<mmax
        mmax = cloud_model.max_azi_num()
        rp0 = (Mbh0+np.sqrt(Mbh0**2-(abh0*Mbh0)**2))
        OmegaBH0 = 0.5*abh0/rp0
        Mir0 = 0.5*rp0
        m = 1
        omega0 = cloud_model.omega_real(m, mu*Mbh0,abh0,0)/Mbh0
        while (not(omega0<m*OmegaBH0)):
            m = m+1
            if (m>mmax): break
            else: omega0 = cloud_model.omega_real(m, mu*Mbh0,abh0,0)/Mbh0
        if (m>mmax): 
            raise ValueError("Error, azimuthal number > %d not supported." % (mmax))

        #Find cloud frequency/mass at saturation
        Jbh0 = abh0*Mbh0**2
        omegaR = 1.0*omega0
        omegaRprevious = 0.0
        rel_omega_tol = 1.0e-10
        max_iter = 100
        i=0
        while (abs(omegaRprevious-omegaR)>rel_omega_tol*mu and i<max_iter):
            omegaRprevious = 1.0*omegaR
            Mbhf = (m**3-np.sqrt(m**6-16.0*m**2*omegaR**2*(m*Mbh0-omegaR*Jbh0)**2))/(8.0*omegaR**2*(m*Mbh0-omegaR*Jbh0))
            Jbhf = Jbh0 -m/omegaR*(Mbh0-Mbhf)
            omegaR = cloud_model.omega_real(m, mu*Mbhf, Jbhf/Mbhf**2, (Mbh0-Mbhf)/Mbhf)/Mbhf
            i = i+1
        if (i>=max_iter):
            warnings.warn(("Saturation condition only satisfied up to relative difference of %e" 
                                 % (abs(omegaRprevious-omegaR)/mu)), RuntimeWarning)

        #Original black hole mass and spin
        self._Mbh0 = Mbh0 
        self._abh0 = abh0 
        # Frequency of cloud initially (at zero mass) 
        self._omegaR0 = omega0
        # Frequency of cloud at saturation
        self._omegaR = omegaR
        #Mass of black hole at saturation
        self._Mbh = Mbhf 
        #Mass of cloud at saturation
        self._Mcloud0 = Mbh0-Mbhf 
        #Dimensionless spin of black hole at saturation
        self._abh = Jbhf/self._Mbh**2
        #Azimuthal number of cloud
        self._m = m
        #Ultralight boson mass
        self._mu = mu
        self._cloud_model = cloud_model
        self._Pgwt = self._cloud_model.power_gw(self._m, self._mu*self._Mbh, self._abh)
        #Spherical harmonic decomposition of strain
        self._hl = self._cloud_model.strain_sph_harm(self._m, self._mu*self._Mbh, self._abh)
        #e-folding time of boson cloud mass
        self._tauI = self._Mbh0/(2*self._cloud_model.omega_imag(self._m, self._mu*self._Mbh0, self._abh0))
    def azimuthal_num(self):
        """Azimuthal number of cloud (mass/gravitational waves have twice this)"""
        return self._m
    def mass_bh_final(self):
        """Black hole mass at saturation"""
        return self._Mbh
    def spin_bh_final(self):
        """Black hole dimensionless spin at saturation"""
        return self._abh
    def efold_time(self):
        """Before saturation, e-folding time of boson cloud mass"""
        return (self._tauI*self._tunit)
    def cloud_growth_time(self):
        """Time for cloud to grow from single boson to saturation"""
        return (self.efold_time()*np.log(self._Mcloud0/(self._mu*self._hbar)))
    def mass_cloud(self, t):
        """Mass of boson cloud as  function of time"""
        t = t/self._tunit
        tnorm = t*self._Pgwt*self._Mcloud0/self._Mbh**2
        H = np.heaviside(tnorm,1.0) 
        treg = (1.0-H)*t #to prevent overflow in exponential
        return (H*self._Mcloud0/(1+tnorm)+(1.0-H)*self._Mcloud0*np.exp(treg/self._tauI))
    def _mass_cloud_dot(self, t):
        """Time derivative of mass of boson cloud as  function of time"""
        t = t/self._tunit
        tnorm = t*self._Pgwt*self._Mcloud0/self._Mbh**2
        H = np.heaviside(tnorm,1.0) 
        treg = (1.0-H)*t #to prevent overflow in exponential
        Mcdot_p = (self._Mcloud0/(1+tnorm))**2*(-1.0*self._Pgwt/self._Mbh**2)
        Mcdot_n = self._Mcloud0*np.exp(treg/self._tauI)/self._tauI
        return (H*Mcdot_p+(1.0-H)*Mcdot_n)
    def power_gw(self, t):
        """Power (luminosity) of gravitational waves as a function of time"""
        Mc = self.mass_cloud(t)
        return (self._Punit*self._Pgwt*(Mc)**2/self._Mbh**2)
    def gw_time(self):
        """Characteristic timescale of GW emission (Mc/P_GW) at saturation"""
        return (self._tunit*self._Mbh**2/(self._Pgwt*self._Mcloud0))
    def freq_gw(self, t):
        """Frequency of gravitational wave signal as a function of time"""
        Mc = self.mass_cloud(t)
        omegaR_vec = np.vectorize(self._cloud_model.omega_real, excluded=[0,1,2])
        H = np.heaviside(t,1.0) 
        fgw = (H*omegaR_vec(self._m, self._mu*self._Mbh, self._abh, Mc/self._Mbh)/self._Mbh
               +(1.0-H)*(self._omegaR+(self._omegaR-self._omegaR0)*(Mc/self._Mcloud0-1)) 
              )/np.pi
        return (fgw/self._tunit)
    def freqdot_gw(self, t):
        """Time derivative of frequency of gravitational wave signal as a function of time"""
        Mc = self.mass_cloud(t)
        Mcdot = self._mass_cloud_dot(t) 
        domegaRdMc_vec = np.vectorize(self._cloud_model.domegar_dmc, excluded=[0,1,2])
        H = np.heaviside(t,1.0) 
        fdotgw = (H*domegaRdMc_vec(self._m, self._mu*self._Mbh, self._abh, Mc/self._Mbh)
                  /self._Mbh**2*Mcdot
                  +(1.0-H)*(self._omegaR-self._omegaR0)*Mcdot/self._Mcloud0)/np.pi
        return (fdotgw/self._tunit**2)
    def strain_char(self, t, dObs=None): 
        """
        A characteristic strain value, defined to be: 
        h0:=(10 P_{GW})^{1/2}/(omega_{GW}*dObs).
        dObs is distance to source.

        In the non-relativistic limit (and for azimuthal_num=1), should have that:
        h_+ = h_0*(1+cos^2(theta))/2*cos(phi(t))
        h_x = h_0*cos(theta)*sin(phi(t))
        """
        if (dObs is None):
            #If not specified, assume strain is evaluated at a distance equal to the black hole mass
            dObs = self._Mbh
        else:
            dObs = dObs/self._dunit
        Mc = self.mass_cloud(t)
        omegaR_vec = np.vectorize(self._cloud_model.omega_real, excluded=[0,1,2])
        omegagw = 2.0*omegaR_vec(self._m, self._mu*self._Mbh, self._abh, Mc/self._Mbh)/self._Mbh
        h0 = (10.0*self._Pgwt*(Mc)**2/self._Mbh**2)**0.5/dObs/omegagw
        return h0 
    def strain_amp(self, t, thetaObs, dObs=None):
        """
        Returns magnitude of two polarizations of strain (h_+,h_x) as a
        function of time and inclination angle with respect to spin axis.  Also
        return delta, where delta is the extra phase difference between
        polarizations.  That is, observed strain is: 
        h_I = F^I_+(t)*h_+(t)cos(phi(t))+F^I_x*h_x(t)*sin(phi(t)+delta) 

        dObs is distance to source.
        """
        if (dObs is None):
            #If not specified, assume strain is evaluated at a distance equal to the black hole mass
            dObs = self._Mbh
        else:
            dObs = dObs/self._dunit
        Mc = self.mass_cloud(t)
        hp = 0.0
        hx = 0.0
        l = 2*self._m 
        for hl0 in self._hl:
            Yp = sYlm(-2,l,2*self._m,thetaObs)
            Ym = sYlm(-2,l,-2*self._m,thetaObs)
            hp = hp + hl0*(Yp+(-1)**l*Ym)
            hx = hx + hl0*(Yp-(-1)**l*Ym)
            l = l + 1
        delta = np.angle(hx)-np.angle(hp)
        #Make delta in [-pi,pi)
        delta = (delta + np.pi) % (2 * np.pi) - np.pi
        hp = np.abs(hp)*(Mc/self._Mbh)/(dObs/self._Mbh)
        hx = np.abs(hx)*(Mc/self._Mbh)/(dObs/self._Mbh)
        return (hp,hx,delta) 
    def phase_gw(self, t):
        """
        Return the gravitational wave phase [phi(t) as defined above] as a
        function of time. Here we assume (take the approximation) that domega/dMc is a
        constant. By convention, the phase is zero at t=0.
        """
        t = t/self._tunit
        tau = self._Mbh**2/(self._Pgwt*self._Mcloud0)
        domegaRdMc_t0 = self._cloud_model.domegar_dmc(self._m, self._mu*self._Mbh, self._abh, 0)/self._Mbh**2 
        omega_M0 = self._cloud_model.omega_real(self._m, self._mu*self._Mbh, self._abh, 0)/self._Mbh
        H = np.heaviside(t,1.0) 
        treg = (1.0-H)*t #to prevent overflow in exponential
        phi = 2.0*(H*(omega_M0*t + domegaRdMc_t0*self._Mcloud0*tau*np.log(1.0+t/tau))
                   +(1.0-H)*(self._omegaR0*t+
                     (self._omegaR-self._omegaR0)*self._tauI*(np.exp(treg/self._tauI)-1.0)))
        return (phi)



"""
If units="physical" units of Msolar=1, and use the following units for
input/output:

mu : electronvolts
Mass : solar mass
time : seconds
frequency : Hz
Power : watts
Distance : Mpc

If units="natural" use units where G=c=hbar=1

If "+alpha" is appended to either "physical" or "natural," then units are the
same as above, except the input mu is taken to be in units of (hbar
c^3)/(G*Mbh0), i.e.  mu is set to the dimensionless "fine structure constant"
alpha. 
"""
def set_units(units, Mbh0):
    if (units=="physical" or units=="physical+alpha"):
        tunit = 4.920551932748678e-06 # (G* solar mass/1 sec)/c^3 
        Punit = 3.6283745e52 # c^5/G in Watts 
        dunit = 4.78691895e-20 # (G* solar mass)/(Mpc *c^2)
        hbar = 1.19727031e-76 # (hbar*c/G)/(solar mass)^2 
        if (units=="physical+alpha"):
            muunit = 1.0/Mbh0
        else:
            muunit = 7.48548859e9 # (eV/hbar)(G* solar mass)/c^3    
    elif (units=="natural" or units=="natural+alpha"):
        tunit = 1.0
        Punit = 1.0
        dunit = 1.0
        hbar = 1.0
        if (units=="natural+alpha"):
            muunit = 1.0/Mbh0
        else:
            muunit = 1.0
    else: 
        raise ValueError("Invalid boson cloud waveform units")
    return (tunit, Punit, dunit, hbar, muunit)



from .cloud_model import CloudModel 
from scipy import interpolate
from scipy import optimize
import numpy as np 
from cmath import sqrt
from pathlib import Path

class RelVector(CloudModel):
    """
    Relativistic (alpha~1) model for vector bosons

    All in inputs are in units where black hole mass=1
    alpha is boson mass (times black hole mass)
    abh is black hole dimensionless spin
    Mcloud is boson cloud mass (as fraction of black hole mass) 
    m is azimuthal number of cloud
    """
    def __init__(self):
        """Caching tabulated and fit data"""

        """ 
        m=1 modes: 
        """
        m1_data = np.load(Path(__file__).parent.joinpath('data/m1_pr_mds.npz'))
        m1_wr = m1_data['wr'].flatten()
        m1_wi = m1_data['wi'].flatten()
        m1_a = m1_data['a'].flatten()
        m1_y = m1_data['y'].flatten()
        m1_dwr = m1_data['dwr'].flatten()
        self._f1wr = interpolate.LinearNDInterpolator(list(zip(m1_y,m1_a)),m1_wr)
        self._f1dwr = interpolate.LinearNDInterpolator(list(zip(m1_y,m1_a)),m1_dwr)
        self._f1wi = interpolate.LinearNDInterpolator(list(zip(m1_y,m1_a)),m1_wi)

        """
        m=2 modes
        """
        m2_data = np.load(Path(__file__).parent.joinpath('data/m2_pr_mds.npz'))
        m2_wr = m2_data['wr'].flatten()
        m2_wi = m2_data['wi'].flatten()
        m2_a = m2_data['a'].flatten()
        m2_y = m2_data['y'].flatten()
        m2_dwr = m2_data['dwr'].flatten()
        self._f2wr = interpolate.LinearNDInterpolator(list(zip(m2_y,m2_a)),m2_wr)
        self._f2dwr = interpolate.LinearNDInterpolator(list(zip(m2_y,m2_a)),m2_dwr)
        self._f2wi = interpolate.LinearNDInterpolator(list(zip(m2_y,m2_a)),m2_wi)

        """
        Fit coefficients
        """
        fit_data = np.load(Path(__file__).parent.joinpath('data/pr_fits.npz'))
        self._amat1 = fit_data['amat1']
        self._bmat1 = fit_data['bmat1']
        self._cmat1 = fit_data['cmat1']
        self._amat2 = fit_data['amat2']
        self._bmat2 = fit_data['bmat2']
        self._cmat2 = fit_data['cmat2']

        """
        Radiation data
        """
        sat_flux_data = np.load(Path(__file__).parent.joinpath('data/pr_sat_gw.npz'))
        m1_flux = sat_flux_data['m1_flux']
        m1_mu = sat_flux_data['m1_mu']
        m1_Z2r = sat_flux_data['m1_z2r']
        m1_Z2i = sat_flux_data['m1_z2i']
        m1_Z3r = sat_flux_data['m1_z3r']
        m1_Z3i = sat_flux_data['m1_z3i']
        m2_flux = sat_flux_data['m2_flux']
        m2_mu = sat_flux_data['m2_mu']
        m2_Z4r = sat_flux_data['m2_z4r']
        m2_Z4i = sat_flux_data['m2_z4i']
        m2_Z5r = sat_flux_data['m2_z5r']
        m2_Z5i = sat_flux_data['m2_z5i']

        self._pwm1 = interpolate.interp1d(m1_mu, m1_flux, kind='cubic')
        self._pwm2 = interpolate.interp1d(m2_mu, m2_flux, kind='cubic')
        self._z2rm1 = interpolate.interp1d(m1_mu, m1_Z2r, kind='cubic')
        self._z2im1 = interpolate.interp1d(m1_mu, m1_Z2i, kind='cubic')
        self._z3rm1 = interpolate.interp1d(m1_mu, m1_Z3r, kind='cubic')
        self._z3im1 = interpolate.interp1d(m1_mu, m1_Z3i, kind='cubic')
        self._z4rm2 = interpolate.interp1d(m2_mu, m2_Z4r, kind='cubic')
        self._z4im2 = interpolate.interp1d(m2_mu, m2_Z4i, kind='cubic')
        self._z5rm2 = interpolate.interp1d(m2_mu, m2_Z5r, kind='cubic')
        self._z5im2 = interpolate.interp1d(m2_mu, m2_Z5i, kind='cubic')

    def max_azi_num(self):
        """Maximum azimuthal number the model is defined for"""
        return 2
    def max_spin(self):
        """Maximum spin the model is defined for"""
        return 0.995
    def omega_real(self, m, alpha, abh, Mcloud):
        """Returns real frequency of cloud oscillation"""
        yl, alphamax, Oh = self._y(m, alpha, abh)
        dwr = self._deltaomega(m, alpha, abh)
        if (m==1):
            if (alpha>=0.05 and abh>=self._aswitch(m)):
                wr = alpha*self._f1wr(yl, abh)
            else:
                # Ensures output fails sup. condition if alpha to large
                if (not(self._maxalphaInterp(m, alpha))):
                    return np.nan
                wr = alpha*(1.0-alpha**2/2.0-35.0*alpha**4/24.0+8.0*alpha**5*abh/3.0)
                wr += alpha*0.9934826313642444*alpha**5*(abh*np.sqrt(1-abh**2)-abh)
                for p in range(6, 9):
                    for q in range(0, 4):
                        wr += alpha**(p+1)*(1.0-abh**2)**(q/2.0)*self._amat1[p-6,q]
            return wr+Mcloud*dwr
        elif (m==2):
            if (alpha>=0.25 and abh>=self._aswitch(m)):
                wr = alpha*self._f2wr(yl, abh)
            else:
                # Ensures output fails sup. condition if alpha to large
                if (not(self._maxalphaInterp(m, alpha))):
                    return np.nan
                wr = alpha*(1.0-alpha**2/8.0-143.0*alpha**4/1920.0+alpha**5*abh/15.0)
                wr += alpha*0.03298155184748759*alpha**5*(abh*np.sqrt(1-abh**2)-abh)
                for p in range(6, 9):
                    for q in range(0, 4):
                        wr += alpha**(p+1)*(1.0-abh**2)**(q/2.0)*self._amat2[p-6,q]
            return wr+Mcloud*dwr
        else:
            raise ValueError("Azimuthal index too large")
    def domegar_dmc(self, m, alpha, abh, Mcloud):
        """Returns derivative of real frequency of cloud w.r.t. cloud mass"""
        return self._deltaomega(m, alpha, abh)
    def omega_imag(self, m, alpha, abh):
        """Returns imaginary frequency, 
           i.e. growth rate of superradiant instability"""
        yl, alphamax, Oh = self._y(m, alpha, abh)
        wr = self.omega_real(m, alpha, abh, 0)
        if (m==1):
            if (alpha>=0.05 and abh>=self._aswitch(m)):
                return -np.exp(self._f1wi(yl, abh))*(wr-m*Oh)
            else:
                wi = 1.0
                fm = 1-abh**2+(abh*m-2.0*wr*(1.0+np.sqrt(1.0-abh**2)))**2
                for p in range(1, 11):
                    for q in range(0, 2):
                        wi += alpha**p*(abh**(q+1)*self._bmat1[q,p-1]+self._cmat1[q,p-1]*(1.0-abh**2)**(q/2.0))
                wi *= -4.0*(1.0+np.sqrt(1.0-abh**2))*fm*alpha**7*(wr-m*Oh)
                return wi
        elif (m==2):
            if (alpha>=0.25 and abh>=self._aswitch(m)):
                return -np.exp(self._f2wi(yl, abh))*(wr-m*Oh)
            else:
                wi = 1.0
                for p in range(5, 11):
                    for q in range(0, 2):
                        wi += alpha**p*(abh**(q+1)*self._bmat2[q,p-5]+self._cmat2[q,p-5]*(1.0-abh**2)**(q/2.0))
                fm = 1.0-abh**2+(abh*m-2.0*wr*(1.0+np.sqrt(1.0-abh**2)))**2
                fm *= 4.0*(1.0-abh**2)+(abh*m-2.0*wr*(1.0+np.sqrt(1.0-abh**2)))**2
                wi *= fm*(1.0+np.sqrt(1.0-abh**2))*alpha**11*(-wr+m*Oh)/864.0
                return wi
        else:
            raise ValueError("Azimuthal index too large")
    def power_gw(self, m, alpha, abh):
        """Returns gravitational wave power, scaled to Mcloud=1 
        Using: omega_R(Mcloud=0), valid only at saturation"""
        if (m==1):
            if (alpha<0.17):
                a10 = 23.6070141940415
                a11 = -115.398015735204
                a12 = 222.729191099650
                return a10*alpha**10+a11*alpha**11+a12*alpha**12
            else:
                return self._pwm1(alpha)
        elif (m==2):
            if (alpha<0.45):
                a14 = 0.00174831340006483
                a15 = -0.00676143664936868
                a16 = 0.00696518854011634
                return a14*alpha**14+a15*alpha**15+a16*alpha**16
            else:
                return self._pwm2(alpha) 
        else:
            raise ValueError("Azimuthal index too large")
    def strain_sph_harm(self, m, alpha, abh):
        """Returns e^{iwt}R h^{\ell 2m} (-2-weighted spherical harmonic components)"""
        spsat = self._spinsat(m, alpha)
        wr = 2.0*self.omega_real(m, alpha, spsat, 0)
        if (m==1):
            if (alpha<0.17):
                z2abs = 2.0*np.pi*np.sqrt(wr**2*self.power_gw(m, alpha, spsat))
                return 2.0*np.array([z2abs,0])/(np.sqrt(2.0*np.pi)*wr**2)
            else:
                z2 = self._z2rm1(alpha)+1j*self._z2im1(alpha)
                z3 = self._z3rm1(alpha)+1j*self._z3im1(alpha)
                return -2.0*np.array([z2, z3])/(np.sqrt(2.0*np.pi)*wr**2)
        elif (m==2):
            if (alpha<0.45):
                z4abs = 2.0*np.pi*np.sqrt(wr**2*self.power_gw(m, alpha, spsat))
                return 2.0*np.array([z4abs,0])/(np.sqrt(2.0*np.pi)*wr**2)
            else:
                z4 = self._z4rm2(alpha)+1j*self._z4im2(alpha)
                z5 = self._z5rm2(alpha)+1j*self._z5im2(alpha)
                return -2.0*np.array([z4, z5])/(np.sqrt(2.0*np.pi)*wr**2)
        else:
            raise ValueError("Azimuthal index too large")
    def _aswitch(self, m):
        return 0.6
    def _maxalphaInterp(self, m, alpha):
        """Returns True if alpha is below the saturation-alpha for each m at abh=aswitch(m)"""
        if (m==1):
            return alpha<0.18
        elif (m==2):
            return alpha<0.35
        else:
            raise ValueError("Azimuthal index too large")
    def _y(self, m, alpha, abh, beta=0.91):
        """Returns utility parameter y, approximate maximal alpha, horizon frequency"""
        if (m==1):
            alpha0 = 0.05
        else:
            alpha0 = 0.25
        Oh = 0.5*abh/(1.0+np.sqrt(1.0-abh**2))
        temp = 9.0*m**3*Oh*beta**2+sqrt(3*m**6*beta**4*(27.0*Oh**2-8.0*beta**2))
        alphamaxC = (-3j+sqrt(3.0))*m**2*beta/(3.0**(5.0/6.0)*temp**(1.0/3.0))\
                        +(1.0+sqrt(3.0)*1j)*temp**(1.0/3.0)/(2.0*3.0**(2.0/3.0)*beta)
        alphamax = alphamaxC.real
        yl = (alpha-alpha0)/(alphamax-alpha0)
        return yl, alphamax, Oh
    def _deltaomega(self,m,alpha,abh):
        """Returns the cloud mass-independent frequency shift due to self-gravity"""
        yl, alphamax, Oh = self._y(m, alpha, abh)
        #Factor of 2 below is to go from gravitational potential energy to frequency shift
        if (m==1):
            if (alpha>=0.05 and abh>=self._aswitch(m)):
                return -2.0*alpha*self._f1dwr(yl, abh)
            else:
                dwr = -5.0*alpha**3/16.0
                dwr += -alpha*(-0.011413831834689237149*alpha**3+0.41745609670046540662*alpha**4-2.4755344267838781391*alpha**5+2.5945305106065412737*alpha**6)
                return 2.0*dwr
        elif (m==2):
            if (alpha>=0.25 and abh>=self._aswitch(m)):
                return -2.0*alpha*self._f2dwr(yl, abh)
            else:
                dwr = -93*alpha**3/1024.0
                dwr += -alpha*(0.0031473331860228763446*alpha**3-0.0062807699451029961463*alpha**4-0.011951847200246173628*alpha**5)
                return 2.0*dwr
        else:
            raise ValueError("Azimuthal index too large")
    def _alphasat(self, m, abh):
        """Bisection to find saturation point, returns alpha at saturation"""
        yh, amaxh, Ohh = self._y(m, 0, abh)
        yl, amaxl, Ohl = self._y(m, 0, abh, 1.0)
        def _sat(al):
            return self.omega_real(m, al, abh, 0)-m*0.5*abh/(1.0+np.sqrt(1.0-abh**2))
        return optimize.bisect(_sat, amaxl, amaxh) 
    def _spinsat(self, m, al):
        """Bisection to find saturation point, returns spin at saturation"""
        def _sat(abh):
            return self.omega_real(m, al, abh, 0)-m*0.5*abh/(1.0+np.sqrt(1.0-abh**2))
        return optimize.bisect(_sat, self.max_spin(), 0)



import unittest
import superrad.nonrel_cloud as nrc
import numpy as np

class TestNonRel(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        """Sets up the cloud model class tested here"""
        self.nrm = [nrc.NonrelScalar(),nrc.NonrelVector()]
        self.maxspin = 1.0
        self.spi = [0.5*self.maxspin, 0.6*self.maxspin, 0.8*self.maxspin, 1.0*self.maxspin]
        self.alp = [0.001, 0.01, 0.1]

    def tearDown(self):
        pass

    def test_non_rel_limit(self):
        """Testing non-rel. limit: alpha=0 and Mcloud=0"""

        for mod in self.nrm:
            """\omega_R"""
            # alpha -> 0, Mcloud = 1
            for i, sp in enumerate(self.spi):
                for j in range(1,mod.max_azi_num()):
                    self.assertEqual(mod.omega_real(j,0,sp,1), 0)
        
            """Mcloud hierarchy"""
            # alpha = 0.1, Mcloud -> 0 vs. alpha = 0.1, Mcloud = 0.1
            for i, sp in enumerate(self.spi):
                for j in range(1,mod.max_azi_num()):
                    self.assertTrue(mod.omega_real(j,0.1,sp,0)>mod.omega_real(j,0.1,sp,0.1))

    def test_delta_omega_consistency(self):
        """Tests self-consistency of frequency drift, derivative of d omega_R/d Mc and deltaoemga()"""
        mcl = [0.001,0.01,0.1]

        for mod in self.nrm:
            for l in range(1,mod.max_azi_num()):
                for i,al in enumerate(self.alp):
                    for j,sp in enumerate(self.spi):
                        for k,mc in enumerate(mcl):
                            wr = mod.omega_real(l,al,sp,mc)
                            dwr = mod.domegar_dmc(l,al,sp,mc)
                            wr -= mc*dwr
                            self.assertEqual(wr, mod.omega_real(l,al,sp,0)) #Mcloud=0


    def test_omega_imag_hierarchy(self):
        """Testing spin hierarchy of \omega_I for fixed alpha"""
        for mod in self.nrm:
            for m in range(1,mod.max_azi_num()):
                for i, al in enumerate(self.alp):
                    res = 0
                    for j, sp in enumerate(self.spi):
                        if (mod.omega_imag(m,al,sp)>res): 
                            res=mod.omega_imag(m,al,sp)
                        else:
                            self.assertTrue(False)

    def test_power_hierarchy(self):
        """Testing azimuthal index hierarchy of power for fixed alpha"""
        for mod in self.nrm:
                for i, al in enumerate(self.alp):
                    res = 0
                    for m in range(mod.max_azi_num(),0,-1):
                        if (mod.power_gw(m,al,0)>res): #power_gw() is spin-independent
                            res=mod.power_gw(m,al,0)
                        else:
                            self.assertTrue(False)

    def test_power_consistency(self):
        """
        Testing power_gw() vs. strain_sph_harm() routines 
        to within tolerance in non-relativistic regimes
        """
        tol = 1e-3
        for mod in self.nrm:
            for m in range(1,mod.max_azi_num()):
                for j, al in enumerate(self.alp):
                    for i, sp in enumerate(self.spi):
                        wr = 2.0*mod.omega_real(m,al,sp,0)
                        pw = mod.power_gw(m,al,sp) 
                        st = mod.strain_sph_harm(m,al,sp)
                        pwst = np.sum(abs(st)**2)*wr**2/(8.0*np.pi)
                        if (abs((pw-pwst)/pw)>tol):
                            self.assertTrue(abs((pw-pwst)/pw)<tol)





from .cloud_model import CloudModel 
from scipy import interpolate
from scipy import optimize
import numpy as np

class NonrelScalar(CloudModel):
    """
    Non-relativistic (alpha<<1) model for scalar bosons

    All in inputs are in units where black hole mass=1
    alpha is boson mass (times black hole mass)
    abh is black hole dimensionless spin
    Mcloud is boson cloud mass (as fraction of black hole mass) 
    m is azimuthal number of cloud
    """
    def max_azi_num(self):
        """Maximum azimuthal number the model is defined for"""
        return 15
    def max_spin(self):
        """Maximum spin the model is defined for"""
        return 1.0
    def _shift_factor(self,m):
        coeffs = [-(93/1024), -(793/18432), -(26333/1048576), -(43191/2621440), -(
  1172755/100663296), -(28539857/3288334336), -(1846943453/
  274877906944), -(14911085359/2783138807808), -(240416274739/
  54975581388800), -(1936010885087/532163627843584), -(62306843256889/
  20266198323167232), -(500960136802799/190277084256403456), -(
  8051112929645937/3530822107858468864), -(21555352563374699/
  10808639105689190400), -(8307059966383480541/4722366482869645213696)
  ]
        return coeffs[m-1]
    def omega_real(self, m, alpha, abh, Mcloud):
        """
        Returns real frequency of cloud oscillation.  For cloud mass
        correction, see eq. A4 in Isi et al.  Phys. Rev. D99,084042 (2019), but
        corrected with extra factor of 2 to go from gravitational energy to
        frequency shift
        """
        #If alpha/m is too high (outside superradiant regime) expansion breaks down
        if (alpha>1.5*m*0.5*abh/(1.0+np.sqrt(1.0-abh**2))):
            return np.nan
        l = m 
        n = l + 1
        fnl = -6/(2*l+1) + 2/n
        hl = 16/(2*l*(2*l+1)*(2*l+2))
        w = (alpha * (1-alpha**2/(2*n**2) - alpha**4/(8*n**4) 
                         + fnl * alpha**4 / (n**3) 
                         + hl * abh * m * alpha**5 /(n**3)))
        if (m <= self.max_azi_num()):
            w+= 2.0*self._shift_factor(m)*alpha**3*Mcloud
        else:
            raise ValueError("Azimuthal index too large")
        return w
    def domegar_dmc(self, m, alpha, abh, Mcloud):
        """Returns derivative of real frequency of cloud w.r.t. cloud mass"""
        dwdMc = 0.0
        if (m <= self.max_azi_num()):
            dwdMc = 2.0*self._shift_factor(m)*alpha**3
        return dwdMc
    def omega_imag(self, m, alpha, abh):
        """Returns imaginary frequency, 
           i.e. growth rate of superradiant instability"""
        omega = self.omega_real(m, alpha, abh,0)
        rplus = 1 + np.sqrt(1-abh**2)
        omegaH = 0.5 * (abh/rplus)
        l = m 
        n = l + 1
        Cnl = 2**(4*l+1) * np.math.factorial(n+l)/(n**(2*l+4)*np.math.factorial(n-l-1)) * (np.math.factorial(l)/(np.math.factorial(2*l) * np.math.factorial(2*l+1)))**2
        glm = 1.0
        for k in range(l):
            glm*= (k+1)**2 *(1-abh**2) + (abh * m - 2*rplus *omega)**2
        return 2 * rplus * Cnl * glm * (m* omegaH - omega)*alpha**(4*l + 5)  
    def power_gw(self, m, alpha, abh):
        """Returns gravitational wave power, scaled to Mcloud=1.
        m=1: Eq. (13) in R. Brito et al. Class. Quantum Grav.32(2015) 134001
        m>1: Eq. (57) in Yoshino et al. PTEP 2014, 943E02 (2014)
        """
        if (m==1):
            return ((484.0+9.0*np.pi**2)/23040.0*alpha**14)
        else: 
            l = m
            n = l+1
            ql = 4*l+10
            cnl = (16**(l+1)*l*(2*l-1)*np.math.factorial(2*l-1-1)**2*np.math.factorial(l+n+1-1)**2)/(n**(4*l+8)*(l+1)*np.math.factorial(l+1-1)**4*np.math.factorial(4*l+3-1)*np.math.factorial(n-l-1)**2)
            return cnl*alpha**ql
    def strain_sph_harm(self, m, alpha, abh):
        omegaGW = 2.0*self.omega_real(m, alpha, abh, 0)
        h22 = -2.0*np.sqrt(2.0*np.pi*self.power_gw(m, alpha, abh))/omegaGW
        return np.array([h22])

class NonrelVector(CloudModel):
    """
    Non-relativistic (alpha<<1) model for vector bosons

    All in inputs are in units where black hole mass=1
    alpha is boson mass (times black hole mass)
    abh is black hole dimensionless spin
    Mcloud is boson cloud mass (as fraction of black hole mass) 
    m is azimuthal number of cloud
    """
    def max_azi_num(self):
        """Maximum azimuthal number the model is defined for"""
        return 5
    def max_spin(self):
        """Maximum spin the model is defined for"""
        return 1.0 
    def _shift_factor(self,m):
        coeffs = [-(5/16), -(93/1024), -(793/18432), -(26333/1048576), -(43191/
  2621440), -(1172755/100663296), -(28539857/3288334336), -(
  1846943453/274877906944), -(14911085359/2783138807808), -(
  240416274739/54975581388800), -(1936010885087/532163627843584), -(
  62306843256889/20266198323167232), -(500960136802799/
  190277084256403456), -(8051112929645937/3530822107858468864), -(
  21555352563374699/10808639105689190400)]
        return coeffs[m-1]
    def omega_real(self, m, alpha, abh, Mcloud):
        """
        Returns real frequency of cloud oscillation.  For cloud mass
        correction, see eq. E4 of Baryakhtar et al.  Phys. Rev. D96,035019
        (2017) but corrected with extra factor of 2 to go from gravitational
        energy to frequency shift, and for m>1 in SuperRad paper
        """
        #If alpha/m is too high (outside superradiant regime) expansion breaks down
        if (alpha>1.5*m*0.5*abh/(1.0+np.sqrt(1.0-abh**2))):
            return np.nan

        # Taking result from Baumann eq.2.30 , dominant mode
        l = m -1
        n = l + 1
        j = l + 1 
        fnlj = -4 * (6*l*j + 3*l + 3*j + 2)/((l+j)*(l+j+1)*(l+j+2)) + 2/n
        hlj = 16/((l+j)*(l+j+1)*(l+j+2))
        # 1/Mbh factor = 1
        w = alpha * ( 1 - alpha**2/(2*n**2) - alpha**4 / (8*n**4) 
                         + fnlj *alpha**4 /(n**3) 
                         + hlj *abh * m * alpha **5 / (n**3) )
        if (m<=self.max_azi_num()):
            w+= 2.0*self._shift_factor(m)*alpha**3*Mcloud
        else:
            raise ValueError("Azimuthal index too large")
        return w
    def domegar_dmc(self, m, alpha, abh, Mcloud):
        """Returns derivative of real frequency of cloud w.r.t. cloud mass"""        
        dwdMc = 0.0
        if (m<=self.max_azi_num()):
            dwdMc = 2.0*self._shift_factor(m)*alpha**3 
        return dwdMc
    def omega_imag(self, m, alpha, abh):
        """Returns imaginary frequency, 
           i.e. growth rate of superradiant instability"""
        omega = self.omega_real(m, alpha, abh,0)
        rplus = 1 + np.sqrt(1-abh**2)
        omegaH = 0.5 * (abh/rplus) 
        l = m -1
        n = l + 1
        j = l + 1 
        Cnlj = 2**(2*l+2*j+1) * np.math.factorial(n+l)/(n**(2*l+4)* np.math.factorial(n-l-1)) * (np.math.factorial(l)/(np.math.factorial(l+j)*np.math.factorial(l+j+1)))**2 * (1+ 2*(1+l-j)*(1-l+j)/(l+j))**2
        gjm = 1.0
        for k in range(j):
            gjm*= (k+1)**2 *(1-abh**2) + (abh * m - 2*rplus *omega)**2
        return 2* rplus *Cnlj * gjm *(m * omegaH - omega) * alpha**(2*l+2*j+5)
    def power_gw(self, m, alpha, abh):
        """Returns gravitational wave power, scaled to Mcloud=1"""
        #Non-relativistic fit for m=1 from Siemonsen+ Phys. Rev. D 101, 024019 (2020)
        #All m>1 (uses flat approximation): Baryakhtar et al. Phys. Rev. D96,035019 (2017) (would give 32/5 or 60 for m=1)
        if (m==1):
            return (16.66*alpha**10)
        elif (m==2):
            return (alpha**14/126.0)
        elif (m==3):
            return 6e-6*alpha**18
        elif (m==4):
            return 2e-9*alpha**22
        elif (m==5):
            return 4e-13*alpha**26
        else:
            raise ValueError("Azimuthal index of power_gw() not consistent with max azi_gw")
    def strain_sph_harm(self, m, alpha, abh):
        omegaGW = 2.0*self.omega_real(m, alpha, abh, 0)
        # only considering dominant l = |m| = 2 mode of radiation
        h22 = -2.0*np.sqrt(2.0*np.pi*self.power_gw(m, alpha, abh))/omegaGW
        return np.array([h22])



import numpy as np
from scipy.integrate import solve_ivp

fields = ["Mc", "Mbh", "Jbh"]
Nfields = len(fields)
d = dict()
for n in range(Nfields):
        d[fields[n]] = n

def EOM(t, y, m, mu, cloud_model):
    """
    Coupled evolution of the boson cloud mass (Mc),
    black hole mass (Mbh), and angular momentum (Jbh).
    """
    ydot = np.zeros_like(y)
    Mc = y[d["Mc"]]
    Mbh = y[d["Mbh"]]
    abh = y[d["Jbh"]]/Mbh**2
    alpha = mu*Mbh
    Pgw = cloud_model.power_gw(m, alpha, abh)*(Mc/Mbh)**2
    omegaR = cloud_model.omega_real(m, alpha, abh, Mc/Mbh)/Mbh
    omegaI = cloud_model.omega_imag(m, alpha, abh)/Mbh
    ydot[d["Mc"]] = 2.0*omegaI*Mc-Pgw
    ydot[d["Mbh"]] = -2.0*omegaI*Mc
    ydot[d["Jbh"]] = -2.0*m/omegaR*omegaI*Mc
    return ydot

class EvolveCloud(object):
    """
    Object for calculating superradiant growth, saturation, and dissipation due to GWs of boson cloud.
    """
    def __init__(self, cloud_model, m, mu, Mbh0, Jbh0, Mc0, t, rtol=1.0e-9):
        #Initial conditions
        y = np.zeros(Nfields)
        y[d["Mbh"]] = Mbh0 
        y[d["Jbh"]] = Jbh0 
        y[d["Mc"]] = Mc0 

        #Integrate ODEs
        sol = solve_ivp(EOM, [t[0],t[-1]], y, args=(m, mu, cloud_model), 
                        method="DOP853", dense_output=True, rtol=rtol, atol=0, max_step=2*(t[1]-t[0]))
        f = sol.sol(t)

        self._m = m
        self._mu = mu
        self._Mc = f[d["Mc"],]
        self._Mbh = f[d["Mbh"],]
        self._abh = f[d["Jbh"],]/f[d["Mbh"],]**2
        alpha = self._Mbh*mu
        Pgw_vec = np.vectorize(cloud_model.power_gw, excluded=[0])
        Pgw = Pgw_vec(m, alpha, self._abh)*(self._Mc/self._Mbh)**2
        omegaI_vec = np.vectorize(cloud_model.omega_imag, excluded=[0])
        omegaI = omegaI_vec(m, alpha, self._abh)/self._Mbh
        self._Mcdot = 2.0*omegaI*self._Mc-Pgw 
    def mass_cloud(self):
        return (self._Mc)
    def mass_cloud_dot(self):
        return (self._Mcdot)
    def mass_bh(self):
        return (self._Mbh)
    def spin_bh(self):
        return (self._abh)




from abc import ABCMeta, abstractmethod

class CloudModel(metaclass=ABCMeta):
    """
    Abstract base class encompassing functions needed
    to compute ultralight boson cloud gravitational waves.

    All inputs are in units where black hole mass=G=c=1
    alpha is boson mass (times black hole mass)
    abh is black hole dimensionless spin
    Mcloud is boson cloud mass (as fraction of black hole mass) 
    m is azimuthal number of cloud
    """
    @abstractmethod
    def max_azi_num(self):
        """Maximum azimuthal number the model is defined for"""
        pass
    @abstractmethod
    def max_spin(self):
        """Maximum spin the model is defined for"""
        pass
    @abstractmethod
    def omega_real(self, m, alpha, abh, Mcloud):
        """Returns real frequency of cloud oscillation"""
        pass
    @abstractmethod
    def domegar_dmc(self, m, alpha, abh, Mcloud):
        """Returns derivative of real frequency of cloud oscillation w.r.t.
        cloud mass: domega_R/dMc"""
        pass
    @abstractmethod
    def omega_imag(self, m, alpha, abh):
        """Returns imaginary frequency, 
           i.e. growth rate of superradiant instability"""
        pass
    @abstractmethod
    def power_gw(self, m, alpha, abh):
        """Returns gravitational wave power, scaled to Mcloud=1"""
        pass
    @abstractmethod
    def strain_sph_harm(self, m, alpha, abh):
        """
        Returns array of C(l=2m,2m), C_(l=2m+1,2m), C_(l=2m+2,2m)...
        corresponding to the spherical harmonics of the gravitational wave strain
        
        Still need to work out conventions here (spin-weighted?, normalization?, what about -2m? etc...)
        """
        pass



import unittest
import numpy as np
import superrad.rel_vec_cloud as rvc
from matplotlib import pyplot as plt
from pathlib import Path

class TestRelVec(unittest.TestCase):

    @classmethod
    def setUpClass(self):    
        """Sets up the vector cloud model class tested here"""
        # PLotting issues if error is found
        self.errpltglb = True

        # Setting up cloud
        self.rv = rvc.RelVector()
        self.mmax = self.rv.max_azi_num()
        self.maxspin = self.rv.max_spin()
        self.spi = [0.5*self.maxspin, 0.6*self.maxspin, 0.8*self.maxspin, 1.0*self.maxspin]
        self.alp = [0.001, 0.01, 0.1]

        # Modes data
        self.m1wr99 = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m1_wr_a0.99.dat'))
        self.m1wi99 = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m1_wi_a0.99.dat'))
        self.m1dwr99 = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m1_dwr_a0.99.dat'))
        self.m2wr99 = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m2_wr_a0.99.dat'))
        self.m2wi99 = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m2_wi_a0.99.dat'))
        self.m2dwr99 = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m2_dwr_a0.99.dat'))
        self.m1wr4 = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m1_wr_a0.4.dat'))
        self.m1wi4 = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m1_wi_a0.4.dat'))
        self.m1dwr4 = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m1_dwr_a0.4.dat'))
        self.m2wr4 = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m2_wr_a0.4.dat'))
        self.m2wi4 = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m2_wi_a0.4.dat'))
        self.m2dwr4 = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m2_dwr_a0.4.dat')) 

        # GW emission data
        self.m1pw = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m1_power_gw.dat'))
        self.m2pw = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m2_power_gw.dat'))
        self.m1strain99 = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m1_strain_a0.99.dat'))
        self.m2strain99 = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_m2_strain_a0.99.dat'))

        # Numerical data
        self.m195num = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_numerical_m1_a0.95.dat'))
        self.m1605num = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_numerical_m1_a0.605.dat'))
        self.m295num = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_numerical_m2_a0.95.dat'))
        self.m2605num = np.loadtxt(Path(__file__).parent.joinpath('./testdata/rel_vec_numerical_m2_a0.605.dat'))

    def tearDown(self):
        plt.show()

    def test_non_rel_limit(self):
        """Testing non-rel. limit: alpha=0 and Mcloud=0"""

        """\omega_R"""
        # alpha -> 0, Mcloud = 1
        for i, sp in enumerate(self.spi):
            # m=1
            self.assertEqual(self.rv.omega_real(1,0,sp,1), 0)
            # m=2
            self.assertEqual(self.rv.omega_real(2,0,sp,1), 0)

        """\Delta\omega_R"""
        # alpha -> 0
        for i, sp in enumerate(self.spi):
            # m=1
            self.assertEqual(self.rv._deltaomega(1,0,sp), 0)
            # m=2
            self.assertEqual(self.rv._deltaomega(2,0,sp), 0)

        """Mcloud hierarchy"""
        # alpha = 0.1, Mcloud -> 0 vs. alpha = 0.1, Mcloud = 0.1
        for i, sp in enumerate(self.spi):
            # m=1
            self.assertTrue(self.rv.omega_real(1,0.1,sp,0)>self.rv.omega_real(1,0.1,sp,0.1))
            # m=2
            self.assertTrue(self.rv.omega_real(2,0.1,sp,0)>self.rv.omega_real(2,0.1,sp,0.1))

    def test_delta_omega_consistency(self):
        """Tests self-consistency of frequency drift, derivative of d omega_R/d Mc and deltaoemga()"""
        mcl = [0.001,0.01,0.1]
        # m=1
        for i,al in enumerate(self.alp):
            for j,sp in enumerate(self.spi):
                for k,mc in enumerate(mcl):
                    wr = self.rv.omega_real(1,al,sp,mc)
                    dwr = self.rv.domegar_dmc(1,al,sp,mc)
                    wr -= mc*dwr
                    self.assertEqual(wr, self.rv.omega_real(1,al,sp,0))
        # m=2
        for i,al in enumerate(self.alp):
            for j,sp in enumerate(self.spi):
                for k,mc in enumerate(mcl):
                    wr = self.rv.omega_real(2,al,sp,mc)
                    dwr = self.rv.domegar_dmc(2,al,sp,mc)
                    wr -= mc*dwr
                    self.assertEqual(wr, self.rv.omega_real(2,al,sp,0))


    def test_omega_imag_hierarchy(self):
        """Testing spin hierarchy of \omega_I(alpha,a) for fixed alpha"""
        for i, al in enumerate(self.alp):
            res1 = 0
            res2 = 0
            for j, sp in enumerate(self.spi):
                # m=1
                if (self.rv.omega_imag(1,al,sp)>res1): res1=self.rv.omega_imag(1,al,sp)
                else: self.assertTrue(False)
                # m=2
                if (self.rv.omega_imag(2,al,sp)>res2): res2=self.rv.omega_imag(2,al,sp)
                else: self.assertTrue(False)

    def test_omega_real_hierarchy(self):
        """Testing spin hierarchy of \omega_R(alpha,a) for fixed alpha"""

        alpl = [0.05,0.06,0.07,0.08,0.09,0.1]
        for i, al in enumerate(alpl):
            res1 = 0
            res2 = 0
            for j, sp in enumerate(self.spi):
                # m=1
                if (self.rv.omega_real(1,al,sp,0)>res1): res1=self.rv.omega_real(1,al,sp,0)
                else: self.assertTrue(False)
                # m=2
                if (self.rv.omega_real(2,al,sp,0)>res2): res2=self.rv.omega_real(2,al,sp,0)
                else: self.assertTrue(False)

    def test_y_utility(self):
        """Testing y utility function"""
        
        """Tesing alphamax and oh"""
        ra1, ro1, ra2, ro2 = 0, 0, 0, 0
        for i, sp in enumerate(self.spi):
            # For fixed alpha = 0.1
            # m=1
            yl1, aml1, ohl1 = self.rv._y(1,0.1,sp)
            # m=2
            yl2, aml2, ohl2 = self.rv._y(2,0.1,sp)
            if (aml1>ra1 and ohl1>ro1 and aml2>ra2 and ohl2>ro2): 
                ra1, ro1, ra2, ro2 = aml1, ohl1, aml2, ohl2
            else:
                self.assertTrue(False)

        """Testing y"""
        ry1, ry2 = 0, 0
        alp1 = [0.05+1e-6,0.25,0.5]
        alp2 = [0.25+1e-6,0.55,0.9]
        for i in range(0,3):
            # For fixed spin = 0.9
            # m=1
            yl1, aml1, ohl1 = self.rv._y(1,alp1[i],0.9)
            # m=2
            yl2, aml2, ohl2 = self.rv._y(2,alp2[i],0.9)
            if (yl1>ry1 and yl2>ry2): 
                ry1, ry2 = yl1, yl2
            else:
                self.assertTrue(False)

    def test_sat_consistency(self):
        """Testing consistency of alphasat() and spinsat() routines to within tolerance"""
        tol = 1e-6
        for i, sp in enumerate(self.spi[:-1]):
            # m=1
            als = self.rv._alphasat(1, sp)
            sps = self.rv._spinsat(1,als)
            self.assertTrue(abs((sps-sp)/sp)<tol)
            # m=2
            als = self.rv._alphasat(2, sp)
            sps = self.rv._spinsat(2,als)
            self.assertTrue(abs((sps-sp)/sp)<tol)

    def test_power_consistency(self):
        """
        Testing power_gw() vs. strain_sph_harm() routines 
        to within tolerance in relativistic regimes:
        The two agree only at the saturated spin for a 
        given alpha (power_gw() assumes saturated state)
        """
        tol = 1e-3

        # m=1
        for j, al in enumerate(self.alp):
            spsat = self.rv._spinsat(1,al)
            wr = 2.0*self.rv.omega_real(1,al,spsat,0)
            pw = self.rv.power_gw(1,al,0) # power_gw is spin-independent
            st = self.rv.strain_sph_harm(1,al,spsat)
            pwst = np.sum(abs(st)**2)*wr**2/(8.0*np.pi)
            if (abs((pw-pwst)/pw)>tol):
                # Plotting power in case difference > tol
                alloc = self.m1pw[:,0]
                for l,alk in enumerate(alloc):
                    spsat = self.rv._spinsat(1,alk)
                    wr = 2.0*self.rv.omega_real(1,alk,spsat,0)
                    pwloc = self.rv.power_gw(1,alk,spsat)
                    stloc = np.sum(abs(self.rv.strain_sph_harm(1,alk,spsat))**2)*wr**2/(8*np.pi)
                    plt.yscale('log')
                    plt.scatter(alk,pwloc,color='C0')
                    plt.scatter(alk,stloc,color='C1')
                self.assertTrue(abs((pw-pwst)/pw)<tol)

        # m=2
        alp2 = [0.05,0.2,0.4]
        for j, al in enumerate(alp2):
            spsat = self.rv._spinsat(1,al)
            wr = 2.0*self.rv.omega_real(2,al,spsat,0)
            pw = self.rv.power_gw(2,al,0) # power_gw is spin-independent
            st = self.rv.strain_sph_harm(2,al,spsat)
            pwst = np.sum(abs(st)**2)*wr**2/(8.0*np.pi)
            if (abs((pw-pwst)/pw)>tol):
                # Plotting power in case difference > tol
                alloc = self.m2pw[:,0]
                for l,alk in enumerate(alloc):
                    spsat = self.rv._spinsat(2,alk)
                    wr = 2.0*self.rv.omega_real(2,alk,spsat,0)
                    pwloc = self.rv.power_gw(2,alk,spsat)
                    stloc = np.sum(abs(self.rv.strain_sph_harm(2,alk,spsat))**2)*wr**2/(8*np.pi)
                    plt.yscale('log')
                    plt.scatter(alk,pwloc,color='C0')
                    plt.scatter(alk,stloc,color='C1')
                self.assertTrue(abs((pw-pwst)/pw)<tol)

    def test_data_interpolation_boundaries(self):
        """
        Tests output of interpolation routines in extrapolation
        regions (=nan, following scipy.interpolate.LinearNDInterpolator() conventions)
        """
        alsw = [0.05,0.25]
        spsw = [self.rv._aswitch(1),self.rv._aswitch(2)]

        # m=1
        for i,al in enumerate(self.alp):
            for j,sp in enumerate(self.spi):
                y, apm, oh = self.rv._y(1,al,sp)
                if (al>alsw[0] and sp>spsw[0]):
                    self.assertFalse(np.isnan(self.rv._f1wr(y, sp)))
                    self.assertFalse(np.isnan(self.rv._f1dwr(y, sp)))
                    self.assertFalse(np.isnan(self.rv._f1wi(y, sp)))
                else:
                    self.assertTrue(np.isnan(self.rv._f1wr(y, sp)))
                    self.assertTrue(np.isnan(self.rv._f1dwr(y, sp)))
                    self.assertTrue(np.isnan(self.rv._f1wi(y, sp)))

        # m=2
        alp2 = [0.1,0.2,0.3,0.5]
        for i,al in enumerate(alp2):
            for j,sp in enumerate(self.spi):
                y, apm, oh = self.rv._y(2,al,sp)
                if (al>alsw[1] and sp>spsw[1]):
                    self.assertFalse(np.isnan(self.rv._f2wr(y, sp)))
                    self.assertFalse(np.isnan(self.rv._f2dwr(y, sp)))
                    self.assertFalse(np.isnan(self.rv._f2wi(y, sp)))
                else:
                    self.assertTrue(np.isnan(self.rv._f2wr(y, sp)))
                    self.assertTrue(np.isnan(self.rv._f2dwr(y, sp)))
                    self.assertTrue(np.isnan(self.rv._f2wi(y, sp)))

    def test_omegaR_against_v1_data(self):
        """Testing \omega_R(alpha,spin) against values obtained for v1 to within tolerance"""
        tol = 1e-6
        errcount = 0
        errplt = self.errpltglb

        # m=1, a=0.99
        altest = self.m1wr99[:,0]
        wrtest = self.m1wr99[:,1]
        for i,wrt in enumerate(wrtest):
            wrc = self.rv.omega_real(1,altest[i],0.99,0)
            if (abs((wrc-wrt)/wrt)>tol):
                if (errplt):
                    plt.plot(altest, wrtest/altest, label='wr_v1, sp=0.99,m=1')
                    wrltest = np.zeros_like(altest)
                    for j,al in enumerate(altest):
                        wrltest[j] = self.rv.omega_real(1,al,0.99,0)
                    plt.plot(altest,wrltest/altest, label, label='wr_SuperRad, sp=0.99,m=1')
                    plt.plot(altest,np.abs((wrtest-wrltest)/wrtest), label='rel_diff, sp=0.99,m=1')
                    plt.yscale('log')
                    plt.xscale('log')
                    plt.legend()
                errcount += 1
                break

        # m=1, a=0.4
        altest = self.m1wr4[:,0]
        wrtest = self.m1wr4[:,1]
        for i,wrt in enumerate(wrtest):
            wrc = self.rv.omega_real(1,altest[i],0.4,0)
            if (abs((wrc-wrt)/wrt)>tol):
                if (errplt):
                    plt.plot(altest, wrtest/altest, label='wr_v1, sp=0.4,m=1')
                    wrltest = np.zeros_like(altest)
                    for j,al in enumerate(altest):
                        wrltest[j] = self.rv.omega_real(1,al,0.4,0)
                    plt.plot(altest,wrltest/altest, label='wr_SuperRad, sp=0.4,m=1')
                    plt.plot(altest,np.abs((wrtest-wrltest)/wrtest), label='rel_diff, sp=0.4,m=1')
                    plt.yscale('log')
                    plt.xscale('log')
                errcount += 1
                break

        # m=2, a=0.99
        altest = self.m2wr99[:,0]
        wrtest = self.m2wr99[:,1]
        for i,wrt in enumerate(wrtest):
            wrc = self.rv.omega_real(2,altest[i],0.99,0)
            if (abs((wrc-wrt)/wrt)>tol):
                if (errplt):
                    plt.plot(altest, wrtest/altest, label='wr_v1, sp=0.99,m=2')
                    wrltest = np.zeros_like(altest)
                    for j,al in enumerate(altest):
                        wrltest[j] = self.rv.omega_real(2,al,0.99,0)
                    plt.plot(altest,wrltest/altest, label='wr_SuperRad, sp=0.99,m=2')
                    plt.plot(altest,np.abs((wrtest-wrltest)/wrtest), label='rel_diff, sp=0.99,m=2')
                    plt.yscale('log')
                    plt.xscale('log')
                errcount += 1
                break

        # m=2, a=0.4
        altest = self.m2wr4[:,0]
        wrtest = self.m2wr4[:,1]
        for i,wrt in enumerate(wrtest):
            wrc = self.rv.omega_real(2,altest[i],0.4,0)
            if (abs((wrc-wrt)/wrt)>tol):
                if (errplt):
                    plt.plot(altest, wrtest/altest, label='wr_v1, sp=0.4,m=2')
                    wrltest = np.zeros_like(altest)
                    for j,al in enumerate(altest):
                        wrltest[j] = self.rv.omega_real(2,al,0.4,0)
                    plt.plot(altest,wrltest/altest, label='wr_SuperRad, sp=0.4,m=2')
                    plt.plot(altest,np.abs((wrtest-wrltest)/wrtest), label='rel_diff, sp=0.4,m=2')
                    plt.yscale('log')
                    plt.xscale('log')
                    plt.legend()
                errcount += 1
                break

        # Raising error if one or more came up
        if (errcount>0):
            print('\n')
            print('Test_omegaR_against_v1_data: Error count = %d/4' % (errcount))
            plt.title('Vector: Testing omega_R (wr) against v1 data')
            plt.legend()
            self.assertTrue(False)

    def test_omegaI_against_v1_data(self):
        """Testing \omega_I(alpha,spin) against values obtained for v1 to within tolerance"""
        tol = 1e-6
        errcount = 0
        errplt = self.errpltglb

        # m=1, a=0.99
        altest = self.m1wi99[:,0]
        witest = self.m1wi99[:,1]
        for i,wit in enumerate(witest):
            wic = self.rv.omega_imag(1,altest[i],0.99)
            if (abs((wic-wit)/wic)>tol): 
                if (errplt):
                    plt.plot(altest, witest, label='wi_v1, sp=0.99,m=1')
                    wiltest = np.zeros_like(altest)
                    for j,al in enumerate(altest):
                        wiltest[j] = self.rv.omega_imag(1,al,0.99)
                    plt.plot(altest,wiltest, label='wi_SuperRad, sp=0.99,m=1')
                    plt.plot(altest,np.abs((witest-wiltest)/witest), label='rel_diff, sp=0.99,m=1')
                    plt.yscale('log')
                    plt.xscale('log')
                errcount += 1
                break

        # m=1, a=0.4
        altest = self.m1wi4[:,0]
        witest = self.m1wi4[:,1]
        for i,wit in enumerate(witest):
            wic = self.rv.omega_imag(1,altest[i],0.4)
            if (abs((wic-wit)/wic)>tol): 
                if (errplt):
                    plt.plot(altest, witest, label='wi_v1, sp=0.4,m=1')
                    wiltest = np.zeros_like(altest)
                    for j,al in enumerate(altest):
                        wiltest[j] = self.rv.omega_imag(1,al,0.4)
                    plt.plot(altest,wiltest, label='wi_SuperRad, sp=0.4,m=1')
                    plt.plot(altest,np.abs((witest-wiltest)/witest), label='rel_diff, sp=0.4,m=1')
                    plt.yscale('log')
                    plt.xscale('log')
                errcount += 1
                break

        # m=2, a=0.99
        altest = self.m2wi99[:,0]
        witest = self.m2wi99[:,1]
        for i,wit in enumerate(witest):
            wic = self.rv.omega_imag(2,altest[i],0.99)
            if (abs((wic-wit)/wic)>tol): 
                if (errplt):
                    plt.plot(altest, witest, label='wi_v1, sp=0.99,m=2')
                    wiltest = np.zeros_like(altest)
                    for j,al in enumerate(altest):
                        wiltest[j] = self.rv.omega_imag(2,al,0.99)
                    plt.plot(altest,wiltest, label='wi_SuperRad, sp=0.99,m=2')
                    plt.plot(altest,np.abs((witest-wiltest)/witest), label='rel_diff, sp=0.99,m=2')
                    plt.yscale('log')
                    plt.xscale('log')
                errcount += 1
                break

        # m=2, a=0.4
        altest = self.m2wi4[:,0]
        witest = self.m2wi4[:,1]
        for i,wit in enumerate(witest):
            wic = self.rv.omega_imag(2,altest[i],0.4)
            if (abs((wic-wit)/wic)>tol): 
                if (errplt):
                    plt.plot(altest, witest, label='wi_v1, sp=0.4,m=2')
                    wiltest = np.zeros_like(altest)
                    for j,al in enumerate(altest):
                        wiltest[j] = self.rv.omega_imag(2,al,0.4)
                    plt.plot(altest,wiltest, label='wi_SuperRad, sp=0.4,m=2')
                    plt.plot(altest,np.abs((witest-wiltest)/witest), label='rel_diff, sp=0.4,m=2')
                    plt.yscale('log')
                    plt.xscale('log')
                errcount += 1
                break

        # Raising error if one or more came up
        if (errcount>0):
            print('\n')
            print('Test_omegaI_against_v1_data: Error count = %d/4' % (errcount))
            plt.title('Vector: Testing omega_I (wi) against v1 data')
            plt.legend()
            self.assertTrue(False)

    def test_modes_against_numerical_data(self):
        """Testing \omega_R(alpha,spin) against values obtained directly from the field's eom"""
        tol = 1e2
        errcount = 0
        errplt = self.errpltglb

        # m\in {1,2}, a\in {0.95,0.605}
        for l in range(0, 2): # m loop
            for k in range(0, 2): # spin loop
                if (k==0 and l==0):
                    altest = self.m195num[:,0]
                    wrtest = self.m195num[:,1]
                    witest = self.m195num[:,2]
                elif (k==1 and l==0):
                    altest = self.m1605num[:,0]
                    wrtest = self.m1605num[:,1]
                    witest = self.m1605num[:,2]
                elif (k==0 and l==1):
                    altest = self.m295num[:,0]
                    wrtest = self.m295num[:,1]
                    witest = self.m295num[:,2]
                elif (k==1 and l==1):
                    altest = self.m2605num[:,0]
                    wrtest = self.m2605num[:,1]
                    witest = self.m2605num[:,2]
                spinin = [0.95,0.605]
                aziin = [1,2]
                for i in range(0,len(wrtest)):
                    wrc = self.rv.omega_real(aziin[l],altest[i],spinin[k],0)
                    wrt = wrtest[i]
                    wic = self.rv.omega_imag(aziin[l],altest[i],spinin[k])
                    wit = witest[i]
                    if (abs((wrc-wrt)/wrt)>tol or abs((wic-wit)/wit)>tol):
                        if (errplt):
                            plt.figure(1)
                            plt.title('Vector: Testing omega_R (wr) against numerical data')
                            plt.plot(altest, wrtest/altest)
                            wrltest = np.zeros_like(altest)
                            for j,al in enumerate(altest):
                                wrltest[j] = self.rv.omega_real(aziin[l],al,spinin[k],0)
                            plt.plot(altest,wrltest/altest, label='wr_SuperRad, sp='+str(spinin[k])+',m='+str(aziin[l]))
                            plt.plot(altest,np.abs((wrtest-wrltest)/wrtest), label='rel_diff, sp='+str(spinin[k])+',m='+str(aziin[l]))
                            plt.yscale('log')
                            plt.xscale('log')
                            plt.legend()
                            plt.figure(2)
                            plt.title('Vector: Testing omega_I (wi) against numerical data')
                            plt.plot(altest, witest)
                            wiltest = np.zeros_like(altest)
                            for j,al in enumerate(altest):
                                wiltest[j] = self.rv.omega_imag(aziin[l],al,spinin[k])
                            plt.plot(altest,wiltest, label='wi_SuperRad, sp='+str(spinin[k])+',m='+str(aziin[l]))
                            plt.plot(altest,np.abs((witest-wiltest)/witest), label='rel_diff, sp='+str(spinin[k])+',m='+str(aziin[l]))
                            plt.yscale('log')
                            plt.xscale('log')
                            plt.legend()
                        errcount += 1
                        break

        # Raising error if one or more came up
        if (errcount>0):
            print('\n')
            print('Test_modes_against_numerical_data: Error count = %d/4' % (errcount))
            self.assertTrue(False)

    def test_deltaomega_against_v1_data(self):
        """Testing \Delta\omega_R(alpha,spin) against values obtained for v1 to within tolerance"""
        tol = 1e-6
        errcount = 0
        errplt = self.errpltglb

        # m=1, a=0.99
        altest = self.m1dwr99[:,0]
        dwrtest = self.m1dwr99[:,1]
        for i,dwrt in enumerate(dwrtest):
            dwrc = self.rv._deltaomega(1,altest[i],0.99)
            if (abs((dwrc-dwrt)/dwrc)>tol):
                if (errplt):
                    plt.plot(altest, -dwrtest, label='dwr_v1, sp=0.99,m=1')
                    dwrltest = np.zeros_like(altest)
                    for j,al in enumerate(altest):
                        dwrltest[j] = self.rv._deltaomega(1,al,0.99)
                    plt.plot(altest,-dwrltest, label='dwr_SuperRad, sp=0.99,m=1')
                    plt.plot(altest,np.abs((dwrtest-dwrltest)/dwrtest), label='rel_diff, sp=0.99,m=1')
                    plt.yscale('log')
                    plt.xscale('log')
                errcount += 1
                break

        # m=1, a=0.4
        altest = self.m1dwr4[:,0]
        dwrtest = self.m1dwr4[:,1]
        for i,dwrt in enumerate(dwrtest):
            dwrc = self.rv._deltaomega(1,altest[i],0.4)
            if (abs((dwrc-dwrt)/dwrc)>tol):
                if (errplt):
                    plt.plot(altest, -dwrtest, label='dwr_v1, sp=0.4,m=1')
                    dwrltest = np.zeros_like(altest)
                    for j,al in enumerate(altest):
                        dwrltest[j] = self.rv._deltaomega(1,al,0.4)
                    plt.plot(altest,-dwrltest, label='dwr_SuperRad, sp=0.4,m=1')
                    plt.plot(altest,np.abs((dwrtest-dwrltest)/dwrtest), label='rel_diff, sp=0.4,m=1')
                    plt.yscale('log')
                    plt.xscale('log')
                errcount += 1
                break

        # m=2, a=0.99
        altest = self.m2dwr99[:,0]
        dwrtest = self.m2dwr99[:,1]
        for i,dwrt in enumerate(dwrtest):
            dwrc = self.rv._deltaomega(2,altest[i],0.99)
            if (abs((dwrc-dwrt)/dwrc)>tol):
                if (errplt):
                    plt.plot(altest, -dwrtest, label='dwr_v1, sp=0.99,m=2')
                    dwrltest = np.zeros_like(altest)
                    for j,al in enumerate(altest):
                        dwrltest[j] = self.rv._deltaomega(2,al,0.99)
                    plt.plot(altest,-dwrltest, label='dwr_SuperRad, sp=0.99,m=2')
                    plt.plot(altest,np.abs((dwrtest-dwrltest)/dwrtest), label='rel_diff, sp=0.99,m=2')
                    plt.yscale('log')
                    plt.xscale('log')
                errcount += 1
                break

        # m=2, a=0.4
        altest = self.m2dwr4[:,0]
        dwrtest = self.m2dwr4[:,1]
        for i,dwrt in enumerate(dwrtest):
            dwrc = self.rv._deltaomega(2,altest[i],0.4)
            if (abs((dwrc-dwrt)/dwrc)>tol):
                if (errplt):
                    plt.plot(altest, -dwrtest, label='dwr_v1, sp=0.4,m=2')
                    dwrltest = np.zeros_like(altest)
                    for j,al in enumerate(altest):
                        dwrltest[j] = self.rv._deltaomega(2,al,0.4)
                    plt.plot(altest,-dwrltest, label='dwr_SuperRad, sp=0.4,m=2')
                    plt.plot(altest,np.abs((dwrtest-dwrltest)/dwrtest), label='rel_diff, sp=0.4,m=2')
                    plt.yscale('log')
                    plt.xscale('log')
                errcount += 1
                break

        # Raising error if one or more came up
        if (errcount>0):
            print('\n')
            print('Test_deltaomega_against_v1_data: Error count = %d/4' % (errcount))
            plt.title('Vector: Testing delta_omega_R (dwr) against v1 data')
            plt.legend()
            self.assertTrue(False)

    def test_power_against_v1_data(self):
        """Testing \dot{E}(alpha,spin=spin_sat) against values obtained for v1 to within tolerance"""
        tol = 1e-6

        # m=1, spin-independent
        altest = self.m1pw[:,0]
        pwtest = self.m1pw[:,1]
        for i,pwt in enumerate(pwtest):
            pwc = self.rv.power_gw(1,altest[i],0)
            if (abs((pwc-pwt)/pwc)>tol): self.assertTrue(False)

        # m=2, spin-independent
        altest = self.m2pw[:,0]
        pwtest = self.m2pw[:,1]
        for i,pwt in enumerate(pwtest):
            pwc = self.rv.power_gw(2,altest[i],0)
            if (abs((pwc-pwt)/pwc)>tol): self.assertTrue(False)

    def test_strain_against_v1_data(self):
        """Testing h^{lm} against values obtained for v1 to within tolerance"""
        tol = 1e-6
        errcount = 0
        errplt = self.errpltglb

        # m=1, a=0.99
        altest = self.m1strain99[:,0]
        strtest = self.m1strain99[:,1]
        for i,strt in enumerate(strtest):
            strc = np.sum(abs(self.rv.strain_sph_harm(1,altest[i],0.99))**2)
            if (abs((strc-strt)/strc)>tol):
                if (errplt):
                    plt.plot(altest, strtest)
                    strltest = np.zeros_like(altest)
                    for j,al in enumerate(altest):
                        strltest[j] = np.sum(abs(self.rv.strain_sph_harm(1,al,0.99))**2)
                    plt.plot(altest,strltest)
                    plt.plot(altest,np.abs((strtest-strltest)/strtest))
                    plt.yscale('log')
                    plt.xscale('log')
                errcount += 1
                break

        # m=2, a=0.99
        altest = self.m2strain99[:,0]
        strtest = self.m2strain99[:,1]
        for i,strt in enumerate(strtest):
            strc = np.sum(abs(self.rv.strain_sph_harm(2,altest[i],0.99))**2)
            if (abs((strc-strt)/strc)>tol):
                if (errplt):
                    plt.plot(altest, strtest)
                    strltest = np.zeros_like(altest)
                    for j,al in enumerate(altest):
                        strltest[j] = np.sum(abs(self.rv.strain_sph_harm(2,al,0.99))**2)
                    plt.plot(altest,strltest)
                    plt.plot(altest,np.abs((strtest-strltest)/strtest))
                    plt.yscale('log')
                    plt.xscale('log')
                errcount += 1
                break

        # Raising error if one or more came up
        if (errcount>0):
            print('\n')
            print('Test_strain_against_v1_data: Error count = %d/2' % (errcount))
            self.assertTrue(False)

